# -*- coding: utf-8 -*-
import coloredlogs, logging
import click
import yaml

from .samplesheet import sheet
from ..basemask import basemask

log = logging.getLogger(__name__)

__version__ = "5.3.1"


@click.group()
@click.option("-l", "--log-level", default="INFO", envvar="LOGLEVEL")
@click.option("-c", "--config", type=click.File("r"))
@click.version_option(version=__version__, prog_name="demux")
@click.pass_context
def demux(context, log_level, config):
    """Making demuxing easier!"""
    setup_logging(level=log_level)
    # log.info('{}: version {}'.format(__package__, __version__))
    context.obj = yaml.full_load(config) if config else {}
    context.obj["log_level"] = log_level


def setup_logging(level="INFO"):
    root_logger = logging.getLogger()
    root_logger.setLevel(level)

    # customize formatter, align each column
    template = "[%(asctime)s] %(name)-25s %(levelname)-8s %(message)s"
    formatter = logging.Formatter(template)

    # add a basic STDERR handler to the logger
    console = logging.StreamHandler()
    console.setLevel(level)
    console.setFormatter(formatter)

    root_logger.addHandler(console)
    return root_logger


demux.add_command(sheet)
demux.add_command(basemask)



# -*- coding: utf-8 -*-

from .samplesheet import (
    Samplesheet,
    HiSeqXSamplesheet,
    NIPTSamplesheet,
    HiSeq2500Samplesheet,
    MiseqSamplesheet,
    SampleSheetValidationException,
    iseqSampleSheet,
)



""" CLI points for samplesheeet action """

import sys
import logging
import copy
import csv
import os

import click

from cglims.api import ClinicalLims, ClinicalSample
from ..utils import (
    Samplesheet,
    HiSeqXSamplesheet,
    NIPTSamplesheet,
    HiSeq2500Samplesheet,
    MiseqSamplesheet,
)

log = logging.getLogger(__name__)


@click.group()
def sheet():
    """Samplesheet commands"""
    pass


@sheet.command()
@click.argument("samplesheet")
@click.option(
    "-a",
    "--application",
    type=click.Choice(["wgs", "wes", "nipt", "miseq"]),
    help="sequencing type",
)
def validate(samplesheet, application):
    """validate a samplesheet"""
    if application == "nipt":
        NIPTSamplesheet(samplesheet).validate()
    elif application == "wes":
        HiSeq2500Samplesheet(samplesheet).validate()
    elif application == "miseq":
        MiseqSamplesheet(samplesheet).validate()
    elif application == "wgs":
        HiSeqXSamplesheet(samplesheet).validate()


@sheet.command()
@click.argument("samplesheet")
def massage(samplesheet):
    """create a NIPT ready SampleSheet"""
    click.echo(NIPTSamplesheet(samplesheet).massage())


@sheet.command()
@click.argument("samplesheet")
@click.option(
    "-a", "--application", type=click.Choice(["miseq", "nipt"]), help="sequencing type"
)
@click.option("-f", "--flowcell", help="for miseq, please provide a flowcell id")
def demux(samplesheet, application, flowcell):
    if application == "nipt":
        """convert NIPT samplesheet to demux'able samplesheet """
        click.echo(NIPTSamplesheet(samplesheet).to_demux())
    elif application == "miseq":
        """convert MiSeq samplesheet to demux'able samplesheet """
        click.echo(MiseqSamplesheet(samplesheet, flowcell).to_demux())
    else:
        log.error("no application provided!")
        sys.exit(1)


@sheet.command()
@click.argument("flowcell")
@click.option(
    "-a",
    "--application",
    type=click.Choice(["wgs", "wes", "nova", "iseq"]),
    help="application type",
)
@click.option(
    "-i",
    "--dualindex",
    is_flag=True,
    default=False,
    help="X: force dual index, not used \
              for NovaSeq!",
)
@click.option(
    "-l",
    "--indexlength",
    default=None,
    help="2500 and NovaSeq: only return this index length",
)
@click.option(
    "-L", "--longest", is_flag=True, help="2500 and NovaSeq: only return longest index"
)
@click.option(
    "-S",
    "--shortest",
    is_flag=True,
    help="2500 and NovaSeq: only return shortest index",
)
@click.option(
    "-d", "--delimiter", default=",", show_default=True, help="column delimiter"
)
@click.option("-e", "--end", default="\n", show_default=True, help="line delimiter")
@click.option(
    "-p",
    "--pad",
    is_flag=True,
    default=False,
    help="add 2 bases to indices with length 8",
)
@click.pass_context
def fetch(
    context,
    flowcell,
    application,
    dualindex,
    indexlength,
    longest,
    shortest,
    pad,
    delimiter=",",
    end="\n",
):
    """
    Fetch a samplesheet from LIMS.
    If a flowcell has dual indices of length 10+10 bp (dual 10) and/or 8+8 bp (dual 8), use
    the option -p, or --pad to add two bases to length 8 indices (AT for index1, AC for index2).
    This will ensure that all indices in the sample sheet are of the same length, namely 10.
    """

    def reverse_complement(dna):
        complement = {"A": "T", "C": "G", "G": "C", "T": "A"}
        return "".join([complement[base] for base in dna[::-1]])

    def get_project(project):
        """ Only keeps the first part of the project name"""
        return project.split(" ")[0]

    lims_api = ClinicalLims(**context.obj["lims"])
    raw_samplesheet = list(lims_api.samplesheet(flowcell))

    if len(raw_samplesheet) == 0:
        sys.stderr.write(f"Samplesheet for {flowcell} not found in LIMS! ")
        context.abort()

    if longest:
        longest_row = max(
            raw_samplesheet, key=lambda x: len(x["index"].replace("-", ""))
        )
        indexlength = len(longest_row["index"].replace("-", ""))

    if shortest:
        shortest_row = min(
            raw_samplesheet, key=lambda x: len(x["index"].replace("-", ""))
        )
        indexlength = len(shortest_row["index"].replace("-", ""))

    # ... fix some 2500 specifics
    if application == "wes":
        # this is how the data is keyed when it gets back from LIMS
        lims_keys = [
            "fcid",
            "lane",
            "sample_id",
            "sample_ref",
            "index",
            "description",
            "control",
            "recipe",
            "operator",
            "project",
        ]
        header = [HiSeq2500Samplesheet.header_map[head] for head in lims_keys]

        if indexlength:
            raw_samplesheet = [
                line
                for line in raw_samplesheet
                if len(line["index"].replace("-", "")) == int(indexlength)
            ]
        for line in raw_samplesheet:
            line["description"] = line["sample_id"]

    # ... fix some X specifics
    if application == "wgs":
        if dualindex:
            lims_keys = [
                "fcid",
                "lane",
                "sample_id",
                "sample_ref",
                "index",
                "index2",
                "sample_name",
                "control",
                "recipe",
                "operator",
                "project",
            ]
            for line in raw_samplesheet:
                line["index2"] = ""
        else:
            lims_keys = [
                "fcid",
                "lane",
                "sample_id",
                "sample_ref",
                "index",
                "sample_name",
                "control",
                "recipe",
                "operator",
                "project",
            ]

        header = [Samplesheet.header_map[head] for head in lims_keys]

        # first do some 10X magic, if any
        new_samplesheet = []
        for i, line in enumerate(raw_samplesheet):
            index = line["index"]
            if len(index.split("-")) == 4:
                for tenx_index in index.split("-"):
                    tenx_line = copy.deepcopy(line)
                    tenx_line["sample_id"] = "{}_{}".format(
                        line["sample_id"], tenx_index
                    )
                    tenx_line["index"] = tenx_index
                    new_samplesheet.append(tenx_line)
            else:
                new_samplesheet.append(line)
        raw_samplesheet = new_samplesheet

        # do some single/dual index stuff
        for i, line in enumerate(raw_samplesheet):
            if not dualindex:
                index = line["index"].split("-")[0]
                raw_samplesheet[i]["index"] = index
                raw_samplesheet[i]["sample_id"] = "{}_{}".format(
                    line["sample_id"], index
                )
            else:
                ori_index = line["index"]
                indexes = ori_index.split("-")
                if len(indexes) == 2:
                    (index1, index2) = indexes
                    raw_samplesheet[i]["index"] = index1
                    raw_samplesheet[i]["index2"] = reverse_complement(index2)
                    raw_samplesheet[i]["sample_id"] = "{}_{}".format(
                        line["sample_id"], ori_index
                    )

        # add [section] header
        click.echo("[Data]")

    if application == "nova":
        if dualindex:
            click.echo(
                click.style(
                    f"No need to specify dual or single index for NovaSeq sample "
                    f"sheets, please use --shortest, --longest, or --indexlength "
                    f"only!",
                    fg="red",
                )
            )
            context.abort()

        if pad and not indexlength:
            click.echo(
                click.style(
                    f"Please specify an index length when using the pad option!"
                    f"Use --longest or --indexlength",
                    fg="red",
                )
            )
            context.abort()

        lims_keys = [
            "fcid",
            "lane",
            "sample_id",
            "sample_ref",
            "index",
            "index2",
            "sample_name",
            "control",
            "recipe",
            "operator",
            "project",
        ]
        header = [Samplesheet.header_map[head] for head in lims_keys]

        path = os.path.abspath(__file__)
        dir_path = os.path.dirname(path)
        # with open(f"{dir_path}/../../files/20181012_Indices.csv") as csv_file:
        with open(
            f"/home/hiseq.clinical/SCRIPTS/git/demultiplexing/files/20181012_Indices.csv"
        ) as csv_file:
            dummy_samples_csv = csv.reader(csv_file, delimiter=",")
            dummy_samples = [row for row in dummy_samples_csv]
            added_dummy_samples = []
            lanes = set([flowcell["lane"] for flowcell in raw_samplesheet])
            flowcell_id = raw_samplesheet[0]["fcid"]

            for lane in lanes:
                sample_indexes = [
                    sample["index"]
                    for sample in raw_samplesheet
                    if sample["lane"] == lane
                ]
                for name, dummy_index in dummy_samples:
                    if not (
                        any(
                            sample_index.startswith(dummy_index)
                            for sample_index in sample_indexes
                        )
                    ):
                        add_dummy_sample = {
                            "control": "N",
                            "description": "",
                            "fcid": flowcell_id,
                            "index": dummy_index,
                            "index2": "",
                            "lane": lane,
                            "operator": "script",
                            "project": "indexcheck",
                            "recipe": "R1",
                            "sample_id": name.replace(" ", "-")
                            .replace("(", "-")
                            .replace(")", "-"),
                            "sample_name": "indexcheck",
                            "sample_ref": "hg19",
                        }

                        added_dummy_samples.append(add_dummy_sample)

            raw_samplesheet.extend(added_dummy_samples)

        if indexlength:
            if pad and int(indexlength) in (16, 20):
                raw_samplesheet = [
                    line
                    for line in raw_samplesheet
                    if len(line["index"].replace("-", "")) in (16, int(indexlength))
                ]
            else:
                raw_samplesheet = [
                    line
                    for line in raw_samplesheet
                    if len(line["index"].replace("-", "")) == int(indexlength)
                ]

        for line in raw_samplesheet:
            if "-" in line["index"]:
                index1, index2 = line["index"].split("-")
                if pad and len(index1) == 8:
                    index1 += "AT"
                    index2 += "AC"
                line["index"] = index1
                line["index2"] = index2
            else:
                if pad and len(line["index"]) == 8:
                    line["index"] += "AT"
                line["index2"] = ""

        # add [section] header
        click.echo("[Data]")

    if application == "iseq":
        if dualindex:
            click.echo(
                click.style(
                    f"No need to specify dual or single index for iSeq sample "
                    f"sheets, please use --shortest, --longest, or --indexlength "
                    f"only!",
                    fg="red",
                )
            )
            context.abort()

        if pad and not indexlength:
            click.echo(
                click.style(
                    f"Please specify an index length when using the pad option!"
                    f"Use --longest or --indexlength",
                    fg="red",
                )
            )
            context.abort()

        lims_keys = [
            "fcid",
            "sample_id",
            "sample_id",
            "sample_id",
            "index",
            "index2",
            "sample_name",
        ]

        header = [
            "FCID",
            "Sample_ID",
            "Sample_Name",
            "Description",
            "index",
            "index2",
            "Sample_Project",
        ]

        if indexlength:
            if pad and int(indexlength) in (16, 20):
                raw_samplesheet = [
                    line
                    for line in raw_samplesheet
                    if len(line["index"].replace("-", "")) in (16, int(indexlength))
                ]
            else:
                raw_samplesheet = [
                    line
                    for line in raw_samplesheet
                    if len(line["index"].replace("-", "")) == int(indexlength)
                ]

        for line in raw_samplesheet:
            if "-" in line["index"]:
                index1, index2 = line["index"].split("-")
                if pad and len(index1) == 8:
                    index1 += "AT"
                    index2 = "AC" + index2
                line["index"] = index1
                line["index2"] = reverse_complement(index2)
            else:
                if pad and len(line["index"]) == 8:
                    line["index"] += "AT"
                line["index2"] = ""

        # add [section] header
        click.echo("[Header]")
        click.echo("[Reads]")
        click.echo("30")
        click.echo("[Data]")

    click.echo(delimiter.join(header))
    for line in raw_samplesheet:
        # fix the project content
        project = get_project(line["project"])
        line["project"] = project
        line["sample_name"] = project

        # print it!
        click.echo(delimiter.join([str(line[lims_key]) for lims_key in lims_keys]))



""" create basemask for demultiplexing based on run parameters """

import logging
import xml.etree.cElementTree as xml_etree

import click

from pathlib import Path

from .utils import (
    HiSeq2500Samplesheet,
    HiSeqXSamplesheet,
    MiseqSamplesheet,
    NIPTSamplesheet,
)

EMPTY_STRING = ""
LOG = logging.getLogger(__name__)


@click.group()
def basemask():
    """Samplesheet commands"""


@basemask.command()
@click.argument("rundir")
@click.option("-l", "--lane", help="lane number")
@click.option(
    "-a",
    "--application",
    type=click.Choice(["miseq", "nipt", "nova", "wes", "wgs"]),
    help="sequencing type",
)
def create(rundir, lane, application):
    """Create a basemask based on SampleSheet.csv and runParameters"""

    # runParameters.xml
    def parse_run_parameters(run_parameters_file):
        """ parse the run parameters file """
        return xml_etree.parse(Path(rundir).joinpath(run_parameters_file))

    def create_basemask(sheet):
        """ create the bcl2fastq basemask """
        run_parameters_file = "runParameters.xml"
        run_params_tree = parse_run_parameters(run_parameters_file)
        read1_len = int(run_params_tree.findtext("Setup/IndexRead1"))
        read2_len = int(run_params_tree.findtext("Setup/IndexRead2"))

        lines = [line for line in sheet.lines_per_column("lane", lane)]

        # get the index lengths
        index1 = lines[0]["index"]
        index2 = lines[0]["index2"] if "index2" in lines[0] else EMPTY_STRING

        # index1 basemask
        index1_n = "n" * (read1_len - len(index1))
        basemask_index1 = "I" + str(len(index1)) + index1_n

        # index2 basemask
        if read2_len == 0:
            click.echo(f"Y151,{basemask_index1},Y151")
        else:
            index2_n = "n" * (read2_len - len(index2))
            if len(index2) > 0:
                basemask_index2 = "I" + str(len(index2)) + index2_n
            else:
                basemask_index2 = index2_n
            click.echo(f"Y151,{basemask_index1},{basemask_index2},Y151")

    def create_novaseq_basemask():
        """ create the bcl2fastq basemask for novaseq flowcells"""

        run_parameters_file = "RunParameters.xml"
        run_params_tree = parse_run_parameters(run_parameters_file)

        indexread1 = int(run_params_tree.findtext("IndexRead1NumberOfCycles"))
        indexread2 = int(run_params_tree.findtext("IndexRead2NumberOfCycles"))
        read1 = int(run_params_tree.findtext("Read1NumberOfCycles"))
        read2 = int(run_params_tree.findtext("Read2NumberOfCycles"))

        novaseq_basemask = f"Y{read1},I{indexread1},I{indexread2},Y{read2}"

        click.echo(f"{novaseq_basemask}")

    def get_application_sheet(application):
        """ parse the samplesheet in the runs directory, based on the type of application """
        sheet = None
        samplesheet = Path(rundir).joinpath("SampleSheet.csv")
        sheet_map = {
            "nipt": NIPTSamplesheet,
            "wes": HiSeq2500Samplesheet,
            "miseq": MiseqSamplesheet,
            "wgs": HiSeqXSamplesheet,
        }
        sheet = sheet_map[application](samplesheet)

        return sheet

    def create_application_basemask(application):
        """ determine the basemask """
        if application == "nova":
            create_novaseq_basemask()
        else:
            create_basemask(get_application_sheet(application))

    create_application_basemask(application)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['demux', 'demux.cli', 'demux.utils']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=5.3.1,<6.0.0',
 'cglims>=1.3.4,<2.0.0',
 'click>=6.7.0,<7.0.0',
 'path.py>=12.5.0,<13.0.0',
 'psutil>=5.7.3,<6.0.0',
 'six>=1.15.0,<2.0.0']

setup_kwargs = {
    'name': 'demux',
    'version': '0.1.0',
    'description': 'Application for demultiplexing sequence data',
    'long_description': None,
    'author': 'henrikstranneheim',
    'author_email': 'henrik.stranneheim@scilifelab.se',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import re
from copy import deepcopy
from collections import OrderedDict
from pathlib import Path


class SampleSheetValidationException(Exception):
    def __init__(self, section, msg, line_nr):
        self.section = section
        self.msg = msg
        self.line_nr = line_nr

    def __str__(self):
        return repr(
            "Section '{}', Line '{}': {}".format(self.section, self.msg, self.line_nr)
        )


class SampleSheetParsexception(Exception):
    pass


class Line(dict):
    @staticmethod
    def _reverse_complement(dna):
        complement = {"A": "T", "C": "G", "G": "C", "T": "A"}
        return "".join([complement[base] for base in dna[::-1]])

    @property
    def dualindex(self, delim="-", revcomp=False):
        if "index2" in self and len(self["index2"]):
            index2 = (
                self._reverse_complement(self["index2"]) if revcomp else self["index2"]
            )
            return self["index"] + delim + index2
        return self["index"]


class Samplesheet(object):
    """SampleSheet.

    Stores the samplesheet in sections: self.section
    e.g. [Data] section will be stored in self.section['[Data]']

    The [Data] section is the actual samplesheet. It consists of a '[Data]' section marker,
    a column header, and the rows of samplesheet data.

    This will be split on line, with each line turned into a dictionary (column header as keys),
    and stored into self.samplesheet.

    e.g.:

    [Data],,
    Flowcell,SampleID,Index
    HHGGFFSS,ADM1123A1,ACTGACTG

    self.section['[Data]'] = [
        {'FCID': 'HHGGFFSS', 'SampleID': 'ADM1123A1', 'Index': 'ACTGACTG'}
    ]

    The original split line with the section marker, will be stored in self.section_markers['[Data]'] = ['[Data]','','']

    """

    # known sections
    HEADER = "[Header]"
    DATA = "[Data]"

    # for the [Data] section: provide a universal header line.
    # The mapping is like this: universal: expected, e.g. for NIPT we have expected Sample_ID,
    # which is mapped to the more universal 'sample_id': 'sample_id': 'Sample_ID'
    # Universal keys follow the python variable naming rules:
    # lowercase with words separated by underscores as necessary to improve readability.
    # One can make one header map for each samplesheet type.
    header_map = {
        "fcid": "FCID",
        "lane": "Lane",
        "sample_id": "SampleID",
        "sample_ref": "SampleRef",
        "index": "index",
        "index2": "index2",
        "sample_name": "SampleName",
        "control": "Control",
        "recipe": "Recipe",
        "operator": "Operator",
        "project": "Project",
    }

    def _get_flowcell(self):
        for line in self.section[self.DATA]:
            if "Flowcell" in line:
                return line["Flowcell"]
        return None

    def __init__(self, samplesheet_path):
        self.samplesheet_path = samplesheet_path
        self.original_sheet = []  # all lines of hte samplesheet
        self.section_markers = (
            dict()
        )  # [Name]: line; does this section have a named section
        self.parse(samplesheet_path)

    def _get_data_header(self):
        header_r = self._get_data_header_r()
        header_map_r = dict((v, k) for k, v in self.header_map.items())
        header = [header_map_r[k] for k in header_r]

        return header

    def _get_data_header_r(self):
        return self.section[self.DATA][0]

    def _get_header_key(self, key):
        if key not in self.header_map:
            raise KeyError("'{}' not in header_map!".format(key))
        return self.header_map[key]

    def parse(self, samplesheet_path, delim=","):
        """
        Parses a Samplesheet, with their fake csv format.
        Should be instancied with the samplesheet path as an argument.
        Will create a dict for each section. Header: (lines)
        """

        name = "[Data]"
        self.section = OrderedDict()
        with open(samplesheet_path) as csvfile:
            for line in csvfile.readlines():
                line = line.strip()
                line = line.split(delim)
                self.original_sheet.append(line)
                if line[0].startswith("["):
                    name = line[0]
                    self.section_markers[name] = line
                    continue  # skip the actual section header

                if name not in self.section:
                    self.section[name] = []

                self.section[name].append(line)

        if self.DATA not in self.section:
            raise SampleSheetParsexception("No data found!")

        header = self._get_data_header()
        self.samplesheet = [
            Line(dict(zip(header, line))) for line in self.section[self.DATA][1:]
        ]

        header_r = self._get_data_header_r()
        self.samplesheet_r = [
            dict(zip(header_r, line)) for line in self.section[self.DATA][1:]
        ]

    def lines(self):
        """ Yields all lines of the [Data] section. """
        for line in self.samplesheet:
            yield line

    def lines_r(self):
        """ Yields all lines of the [Data] section based on the original header """
        for line in self.samplesheet_r:
            yield line

    def raw(self, delim=",", end="\n"):
        """Reconstructs the sample sheet. """
        rs = []
        for line in self.original_sheet:
            rs.append(delim.join(line))
        return end.join(rs)

    def samples(self, column="sample_id"):
        """ Return all samples in the samplesheet """
        return self.column(column)

    def samples_r(self, column="SampleID"):
        """ Return all samples in the samplesheet based on the original header"""
        return self.column_r(column)

    def column(self, column):
        """ Return all values from a column in the samplesheet """
        for line in self.samplesheet:
            yield line[column]

    def column_r(self, column):
        """ Return all values from a column in the samplesheet based on the original header"""
        for line in self.samplesheet_r:
            yield line[column]

    def cell(self, line, column):
        """ return the contents of a column in a line """

        return line[self._get_header_key(column)]

    def lines_per_column(self, column, content):
        """Return all lines with the same column content
        e.g. return all lines of column='lane' content='1'"""
        for line in self.samplesheet:
            if line[column] == content:
                yield line

    def lines_per_column_r(self, column, content):
        """Return all lines with the same column content
        e.g. return all lines of column='Lane' content='1'"""
        for line in self.samplesheet_r:
            if line[column] == content:
                yield line

    def is_pooled_lane(self, lane, column="lane"):
        """ Return True if lane contains multiple samples """
        lane_count = 0
        lane = str(lane)
        for line in self.samplesheet:
            if line[column] == lane:
                lane_count += 1

            if lane_count > 1:
                return True

        return False

    def is_pooled_lane_r(self, lane, column="lane"):
        """ Return True if lane contains multiple samples based on the orignal header """
        lane_count = 0
        lane = str(lane)
        for line in self.samplesheet_r:
            if line[column] == lane:
                lane_count += 1

            if lane_count > 1:
                return True

        return False

    def validate(self):
        """ General validation of a samplesheet """

        def _validate_length(section):
            if len(section) > 2:
                header = section[0]
                lines = section[1:]
                for i, line in enumerate(lines):
                    if len(header) != len(line):
                        msg = "'{}': #fields != #fields in header".format(line)
                        # add i + 2 as it makes it easier to spot the 'wrong' line
                        return (msg, i + 2)
            return True

        def _validate_uniq_index(samplesheet):
            lanes = list(set(self.column("lane")))
            for lane in lanes:
                if self.is_pooled_lane(lane, column="lane"):
                    sample_of = dict()
                    for line in self.lines_per_column("lane", lane):
                        index = line["index"]
                        if index not in sample_of:
                            sample_of[index] = set()
                        sample_of[index].add(line["sample_id"])

                    for index, samples in sample_of.items():
                        if len(samples) > 1:
                            return (
                                "Same index for {} on lane {}".format(
                                    " , ".join(samples), lane
                                ),
                                index,
                            )

            return True

        def _validate_sample_name(samplesheet):
            for i, line in enumerate(samplesheet):
                forbidden_chars = set(" ")
                if any((c in forbidden_chars) for c in line["sample_id"]):
                    return (
                        "Sample contains forbidden chars ({}): {}".format(
                            forbidden_chars, line["sample_id"]
                        ),
                        i + 2,
                    )

        rs = _validate_uniq_index(self.samplesheet)
        if type(rs) is tuple:
            raise SampleSheetValidationException(self.DATA, rs[1], rs[0])

        for section_marker, section in self.section.items():
            validation_section = section[
                :
            ]  # only validate the content, not the [Data] header
            rs = _validate_length(validation_section)
            if type(rs) is tuple:
                raise SampleSheetValidationException(section_marker, rs[1], rs[0])

        rs = _validate_sample_name(self.samplesheet)
        if type(rs) is tuple:
            raise SampleSheetValidationException(self.DATA, rs[1], rs[0])

        return True


class HiSeqXSamplesheet(Samplesheet):
    def unparse(self, delim=","):
        """Reconstruct the sample sheet based on the (modified) parsed values. """
        rs = []
        yield "[Data]"
        yield delim.join(self._get_data_header_r())
        for line in self.samplesheet:
            line_r = []
            for part in self._get_data_header():
                line_r.append(line[part])
            yield delim.join(line_r)

    def validate(self):
        Samplesheet.validate(self)

        def _validate_project_samplename():
            for i, line in enumerate(self.lines()):
                if line["project"] != line["sample_name"]:
                    # add i + 2 as it makes it easier to spot the 'wrong' line
                    line_nr = i + 2
                    msg = "Project and SampleName cannot be different!"
                    return (msg, line_nr)

        def _validate_index():
            for i, line in enumerate(self.lines()):
                if line["index"] == "":
                    # add i + 2 as it makes it easier to spot the 'wrong' line
                    line_nr = i + 2
                    msg = "Missing index!"
                    return (msg, line_nr)

        for rs in [_validate_index(), _validate_project_samplename()]:
            if type(rs) is tuple:
                raise SampleSheetValidationException(self.DATA, rs[1], rs[0])

        return True


class iseqSampleSheet(Samplesheet):

    header_map = {
        "fcid": "FCID",
        "sample_id": "Sample_ID",
        "sample_name": "Sample_Name",
        "description": "Description",
        "index": "index",
        "index2": "index2",
        "project": "Sample_Project",
    }


class MiseqSamplesheet(Samplesheet):

    header_map = {
        "lane": "Lane",
        "sample_id": "Sample_ID",
        "sample_name": "Sample_Name",
        "sample_plate": "Sample_Plate",
        "sample_well": "Sample_Well",
        "i7_index_id": "I7_Index_ID",
        "index": "index",
        "sample_project": "Sample_Project",
        "index2": "index2",
        "i5_index_id": "I5_Index_ID",
        "genome_folder": "GenomeFolder",
        "description": "Description",
    }

    si5 = {
        "TAGATCGC": "S501",
        "CTCTCTAT": "S502",
        "TATCCTCT": "S503",
        "AGAGTAGA": "S504",
        "GTAAGGAG": "S505",
        "ACTGCATA": "S506",
        "AAGGAGTA": "S507",
        "CTAAGCCT": "S508",
        "CGTCTAAT": "S510",
        "TCTCTCCG": "S511",
        "TCGACTAG": "S513",
        "TTCTAGCT": "S515",
        "CCTAGAGT": "S516",
        "GCGTAAGA": "S517",
        "CTATTAAG": "S518",
        "AAGGCTAT": "S520",
        "GAGCCTTA": "S521",
        "TTATGCGA": "S522"
        #'TAGATCGC': 'N501',
        #'CTCTCTAT': 'N502',
        #'TATCCTCT': 'N503',
        #'AGAGTAGA': 'N504',
        #'GTAAGGAG': 'N505',
        #'ACTGCATA': 'N506',
        #'AAGGAGTA': 'N507',
        #'CTAAGCCT': 'N508',
    }
    ni7 = {
        "TAAGGCGA": "N701",
        "CGTACTAG": "N702",
        "AGGCAGAA": "N703",
        "TCCTGAGC": "N704",
        "GGACTCCT": "N705",
        "TAGGCATG": "N706",
        "CTCTCTAC": "N707",
        "CAGAGAGG": "N708",
        "CGAGGCTG": "N710",
        "AAGAGGCA": "N711",
        "GTAGAGGA": "N712",
        "GCTCATGA": "N714",
        "ATCTCAGG": "N715",
        "ACTCGCTA": "N716",
        "GGAGCTAC": "N718",
        "GCGTAGTA": "N719",
        "CGGAGCCT": "N720",
        "TACGCTGC": "N721",
        "ATGCGCAG": "N722",
        "TAGCGCTC": "N723",
        "ACTGAGCG": "N724",
        "CCTAAGAC": "N726",
        "CGATCAGT": "N727",
        "TGCAGCTA": "N728",
        "TCGACGTC": "N729",
    }
    di7 = {
        "ATTACTCG": "D701",
        "TCCGGAGA": "D702",
        "CGCTCATT": "D703",
        "GAGATTCC": "D704",
        "ATTCAGAA": "D705",
        "GAATTCGT": "D706",
        "CTGAAGCT": "D707",
        "TAATGCGC": "D708",
        "CGGCTATG": "D709",
        "TCCGCGAA": "D710",
        "TCTCGCGC": "D711",
        "AGCGATAG": "D712",
    }
    di5 = {
        "TATAGCCT": "D501",
        "ATAGAGGC": "D502",
        "CCTATCCT": "D503",
        "GGCTCTGA": "D504",
        "AGGCGAAG": "D505",
        "TAATCTTA": "D506",
        "CAGGACGT": "D507",
        "GTACTGAC": "D508",
    }

    def __init__(self, samplesheet_path, flowcell=None, sequencing_date=None):
        Samplesheet.__init__(self, samplesheet_path)
        if flowcell == None:
            flowcell = Path(samplesheet_path).dirname().basename().split("_")[-1]
        if sequencing_date == None:
            sequencing_date = Path(samplesheet_path).dirname().basename().split("_")[0]
        self.flowcell = flowcell
        self.sequencing_date = sequencing_date

    def _get_flowcell(self):
        return self.flowcell

    def to_demux(self, delim=",", end="\n"):
        """ Convert miseq to hiseq style samplesheet for demultiplexing. """

        checked_indexes = {}  # the indexes in the SampleSheet

        def clean(input):
            return re.sub(r"[ _/]+", "", input)

        def check_index(index):
            checked_indexes[index] = 1

        def get_undetermined_indexes():

            # combine the D indexes
            for di7_index, di7_name in self.di7.items():
                for di5_index, di5_name in self.di5.items():
                    d_index = di7_index + "-" + di5_index
                    if d_index not in checked_indexes:
                        yield d_index, str(di7_name + "-" + di5_name)

            # combine the other indexes
            for ni7_index, ni7_name in self.ni7.items():
                for si5_index, si5_name in self.si5.items():
                    ns_index = ni7_index + "-" + si5_index
                    if ns_index not in checked_indexes:
                        yield ns_index, str(ni7_name + "-" + si5_name)

        expected_header = [
            "FCID",
            "Lane",
            "SampleID",
            "SampleRef",
            "Index",
            "Description",
            "Control",
            "Recipe",
            "Operator",
            "SampleProject",
        ]

        # get the experiment name
        flowcell_id = self._get_flowcell()
        cur_date = self.sequencing_date

        header = self.section[self.DATA][0]  # '0' is the csv header
        data_lines = (
            []
        )  # the new data section. Each line holds a dict with the right header keys
        data_lines.append(expected_header)
        for line in self.samplesheet:
            data_line = {}
            data_line["FCID"] = flowcell_id
            data_line["Lane"] = "1"
            data_line["SampleID"] = cur_date + "-" + clean(line["sample_id"])
            data_line["SampleRef"] = "hg19"
            data_line["Index"] = clean(line["index"]) + "-" + clean(line["index2"])
            data_line["Description"] = line["description"]
            data_line["Control"] = "N"
            data_line["Recipe"] = "R1"
            data_line["Operator"] = "MS"
            data_line["SampleProject"] = clean(line["sample_project"])

            check_index(clean(line["index"]) + "-" + clean(line["index2"]))

            ordered_line = []
            for head in expected_header:
                ordered_line.append(data_line[head])
            data_lines.append(ordered_line)

        # add the undetermined indexes
        for undetermined_index, undetermined_index_name in get_undetermined_indexes():
            data_line = {}
            data_line["FCID"] = flowcell_id
            data_line["Lane"] = "1"
            data_line["SampleID"] = cur_date + "-" + undetermined_index_name
            data_line["SampleRef"] = "hg19"
            data_line["Index"] = undetermined_index
            data_line["Description"] = "ctmr"
            data_line["Control"] = "N"
            data_line["Recipe"] = "R1"
            data_line["Operator"] = "script"
            data_line["SampleProject"] = "Undetermined"

            ordered_line = []
            for head in expected_header:
                ordered_line.append(data_line[head])
            data_lines.append(ordered_line)

        rs = []
        for line in data_lines:
            rs.append(delim.join(line))

        return end.join(rs)

    def validate(self):
        Samplesheet.validate(self)

        def _validate_missing_index():
            """We know what indexes to expect in the samplesheet. Check which ones are missing
            or have been added."""

            def is_readymade_index(index):
                for di7_index, di7_name in self.di7.items():
                    for di5_index, di5_name in self.di5.items():
                        d_index = di7_index + "-" + di5_index
                        if index == d_index:
                            return index

                # combine the other indexes
                for ni7_index, ni7_name in self.ni7.items():
                    for si5_index, si5_name in self.si5.items():
                        ns_index = ni7_index + "-" + si5_index
                        if index == ns_index:
                            return index

            for i, line in enumerate(self.lines()):
                readymade_index = is_readymade_index(line["index"])

                if not readymade_index:
                    return (
                        "Index {} not in readymade indexes!".format(line["index"]),
                        i,
                    )

        rs = _validate_missing_index()
        if type(rs) is tuple:
            raise SampleSheetValidationException(self.DATA, rs[1], rs[0])


class HiSeq2500Samplesheet(Samplesheet):

    header_map = {
        "fcid": "FCID",
        "lane": "Lane",
        "sample_id": "SampleID",
        "sample_ref": "SampleRef",
        "index": "Index",
        "sample_name": "SampleName",
        "control": "Control",
        "recipe": "Recipe",
        "operator": "Operator",
        "description": "Description",
        "project": "SampleProject",
    }


class NIPTSamplesheet(Samplesheet):

    header_map = {
        "lane": "Lane",
        "sample_id": "Sample_ID",
        "sample_name": "Sample_Name",
        "sample_plate": "Sample_Plate",
        "sample_well": "Sample_Well",
        "i7_index_id": "I7_Index_ID",
        "index": "index",
        "sample_project": "Sample_Project",
        "description": "Description",
        "sample_type": "SampleType",
        "library_nm": "Library_nM",
    }

    def _get_flowcell(self):
        # get the experiment name
        for line in self.section[self.HEADER]:
            if line[0] == "Experiment Name":
                return line[1]
        return None

    def _get_project_id(self):
        # get the experiment name
        for line in self.section[self.HEADER]:
            if line[0] == "Investigator Name":
                return line[1].split("_")[1]
        return None

    def massage(self, delim=",", end="\n"):
        """Abuses the Investigator Name field to store information about the run.

        Reshuffles the [Data] section so that it becomes a valid sample sheet.

        Returns a massaged SampleSheet.
        """
        # get the experiment name
        flowcell_id = self._get_flowcell()

        section_copy = deepcopy(self.section)

        for i, line in enumerate(section_copy[self.HEADER]):
            if line[0] == "Investigator Name":
                investigator_name = re.split(" |_", line[1])
                if investigator_name[-1] != flowcell_id:
                    investigator_name.append(flowcell_id)
                line[1] = "_".join(investigator_name)
                section_copy[self.HEADER][i] = line

        rs = []
        for section_marker, section in section_copy.items():
            if section_marker in self.section_markers:
                rs.append(delim.join(self.section_markers[section_marker]))
            for line in section:
                rs.append(delim.join(line))
        return end.join(rs)

    def to_demux(self, delim=",", end="\n"):
        """Replaced the [Data] section with a demuxable [Data] section.

        This is non destructive and will only return a demuxable samplesheet.

        Convert the Data section from
            Lane,Sample_ID,Sample_Name,Sample_Plate,Sample_Well,I7_Index_ID,index,Sample_Project,Description,SampleType
        to
            FCID,Lane,SampleID,SampleRef,Index,Description,Control,Recipe,Operator,SampleProject
        """

        expected_header = [
            "FCID",
            "Lane",
            "SampleID",
            "SampleRef",
            "Index",
            "Description",
            "Control",
            "Recipe",
            "Operator",
            "SampleProject",
        ]

        # get the experiment name
        flowcell_id = self._get_flowcell()
        project_id = self._get_project_id()

        header = self.section[self.DATA][0]  # '0' is the csv header
        data_lines = (
            []
        )  # the new data section. Each line holds a dict with the right header keys
        data_lines.append(expected_header)
        for i, line in enumerate(self.section[self.DATA][1:]):
            data_line = dict(zip(header, line))

            data_line["FCID"] = flowcell_id
            data_line["SampleID"] = data_line["Sample_ID"]
            data_line["SampleRef"] = "hg19"
            data_line["Index"] = data_line["index"]
            data_line["Description"] = data_line["SampleType"]
            data_line["Control"] = project_id
            data_line["Recipe"] = "R1"
            data_line["Operator"] = "NN"
            data_line["SampleProject"] = project_id

            ordered_line = []
            for head in expected_header:
                ordered_line.append(data_line[head])
            data_lines.append(ordered_line)

        rs = []
        for line in data_lines:
            rs.append(delim.join(line))
        return end.join(rs)
